import hashlib

from jose.utils import base64url_decode, base64url_encode
from jose import jwk
from Crypto.Signature import PKCS1_PSS
from Crypto.Hash import SHA256
from eth_account.messages import encode_defunct, _hash_eip191_message
from Crypto.PublicKey import RSA

def get_url(server, path):
    if server.endswith('/'):
        return server[:-1] + path
    return server + path

def is_ar_address(address):
    if len(address) != 43:
        return False
    try:
        base64url_decode(address.encode())
    except:
        return False
    return True

def eip191_hash(message):
    return _hash_eip191_message(encode_defunct(text=message))

def verify_ar_sig(owner, message, sig):
    data = {'kty':'RSA', 'e':"AQAB", 'n':owner}
    jwk_ = jwk.construct(data, algorithm=jwk.ALGORITHMS.RS256)
    public_key = RSA.importKey(jwk_.to_pem())
    hash_ = SHA256.new(eip191_hash(message))
    sig = base64url_decode(sig.encode())
    #return PKCS1_PSS.new(public_key, saltLen=32).verify(hash_, sig)
    return PKCS1_PSS.new(public_key).verify(hash_, sig)

def owner_to_address(owner):
    result = base64url_encode(hashlib.sha256(base64url_decode(owner.encode('ascii'))).digest()).decode()
    return result


import json
import random
import time, requests
from .utils import get_url
from .transaction import Transaction
from .token import get_token_tag, get_token_decimal, get_token_id

class Everpay:
    def __init__(self, everpay_server_url):
        self.api_server = everpay_server_url

    def get_info(self):
        url = get_url(self.api_server, '/info')
        return requests.get(url).json()

    def get_balance(self, account, chain_type='', chain_id='', token_symbol='', token_id=''):
        if chain_type and chain_id and token_symbol:
            token_tag = get_token_tag(chain_type, chain_id, token_symbol, token_id)
            path = '/balance/%s/%s' % (token_tag, account.lower())
        else:
            path = '/balances/%s' % account.lower()
        url = get_url(self.api_server, path)
        return requests.get(url).json()

    def get_txs(self, account='', order='desc', page=None):
        path = '/txs'
        if account:
            path = '/txs/%s' % account
        if page:
            path = '%s/?order=%s&page=%i' % (path, order, int(page))
        else:
            path = '%s/?order=%s' % (path, order)
        url = get_url(self.api_server, path)
        return requests.get(url).json()

    def get_tx(self, hash):
        path = '/tx/%s' % hash
        url = get_url(self.api_server, path)
        return requests.get(url).json()


class Account:
    def __init__(self, everpay_server_url, address, private_key, ar_wallet, fee_recipient):
        self.everpay = Everpay(everpay_server_url)
        self.address = address
        self.private_key = private_key
        self.ar_wallet = ar_wallet
        self.fee_recipient = fee_recipient

    def get_balance(self, chain_type, chain_id, token_symbol, token_id=''):
        chain_type = chain_type.lower()
        chain_id = str(chain_id).lower()
        token_symbol = token_symbol.lower()
        return self.everpay.get_balance(self.address, chain_type, chain_id, token_symbol, token_id='')

    def get_txs(self):
        return self.everpay.get_txs(self.address)

    def transfer(self, to, amount, chain_type, chain_id, token_symbol, token_id='', data=''):

        chain_type = chain_type.lower()
        chain_id = str(chain_id).lower()
        token_symbol = token_symbol.lower()

        if not token_id:
            token_id = get_token_id(chain_type, chain_id, token_symbol)

        amount_ = str(amount)
        
        if self.ar_wallet and data:
            data = json.loads(data)
            data['arOwner'] = self.ar_wallet.owner
            data = json.dumps(data)
        if self.ar_wallet and not data:
            data = json.dumps({'arOwner': self.ar_wallet.owner})

        t = Transaction(tx_id='', token_symbol=token_symbol, action='transfer', from_=self.address, to=to,
                        amount=amount_, fee='0', fee_recipient=self.fee_recipient, nonce=str(int(time.time() * 1000)),
                        token_id=token_id, chain_type=chain_type, chain_id=chain_id, data=data, version='v1')
        
        if self.private_key:
            t.sign(self.private_key)
        elif self.ar_wallet:
            t.sign_with_ar_wallet(self.ar_wallet)
        else:
            raise Exception('No key or wallet')

        return t, t.post(self.everpay.api_server).content

    def bundle(self, to, chain_type, chain_id, token_symbol, token_id='', data=''):

        chain_type = chain_type.lower()
        chain_id = str(chain_id).lower()
        token_symbol = token_symbol.lower()

        if not token_id:
            token_id = get_token_id(chain_type, chain_id, token_symbol)

        amount_ = '0'
        
        t = Transaction(tx_id='', token_symbol=token_symbol, action='bundle', from_=self.address, to=to,
                        amount=amount_, fee='0', fee_recipient=self.fee_recipient, nonce=str(int(time.time() * 1000)),
                        token_id=token_id, chain_type=chain_type, chain_id=chain_id, data=data, version='v1')
        t.sign(self.private_key)

        return t, t.post(self.everpay.api_server).content



from jose.utils import base64url_encode
from web3.auto import w3
from eth_account.messages import encode_defunct, _hash_eip191_message
import json
from .utils import *

class BundleItem:
    def __init__(self, token_tag, chain_id, from_, to, amount):
        self.token_tag = token_tag
        self.chain_id = chain_id
        self.from_ = from_
        self.to = to
        self.amount = amount

    def to_dict(self):
        return {
            'tag': self.token_tag,
            'chainID': self.chain_id,
            'from': self.from_,
            'to': self.to,
            'amount': self.amount
        }

class BundleData:
    def __init__(self, item, expiration, salt, version):
        
        self.items = []
        self.add_item(item)

        self.expiration = expiration
        self.salt = salt
        self.version = version
        self.sigs = {}

    def get_data(self):
        data = {
            'bundle': {
                'items': [item.to_dict() for item in self.items],
                'expiration': self.expiration,
                'salt': self.salt,
                'version': self.version,
                'sigs': self.sigs,
            }
        }
        return json.dumps(data, separators=(',', ':'), sort_keys=False)

    def get_data_to_sign(self):
        data = {
                'items': [item.to_dict() for item in self.items],
                'expiration': self.expiration,
                'salt': self.salt,
                'version': self.version,
            }

        return json.dumps(data, separators=(',', ':'), sort_keys=False)

    def get_hash_message(self):
        #print('msg:', self.get_data_to_sign())
        return _hash_eip191_message(encode_defunct(text=self.get_data_to_sign()))
    
    def get_sigs(self):
        return self.sigs

    def add_item(self,item):
        if not isinstance(item, BundleItem):
            raise TypeError('item must be a instance of BundleItem')
        self.items.append(item)

    def add_sig(self, address, sig):
        if is_ar_address(address):
            sig_, owner = sig.split(',')
            
            if owner_to_address(owner) != address:
                raise ValueError('invalid signature: address is not from owner')
            message = self.get_data_to_sign()
            if not verify_ar_sig(owner, message, sig_):
                raise ValueError('invalid signature')

            self.sigs[address] = sig
        else:
            message = encode_defunct(text=self.get_data_to_sign())
            sign_address = w3.eth.account.recover_message(message, signature=sig)

            if sign_address.lower() != address.lower():
                raise ValueError('invalid signature')
            self.sigs[address] = sig

    def sign(self, address, pk=None, wallet=None):
        if is_ar_address(address):
            message = self.get_hash_message()
            sig = base64url_encode(wallet.sign(message)).decode()
            sig = f'{sig},{wallet.owner}'

            self.add_sig(address, sig)
        else:
            pk = bytearray.fromhex(pk)
            message = encode_defunct(text=self.get_data_to_sign())
            sig = w3.eth.account.sign_message(message, private_key=pk)
            self.add_sig(address, sig.signature.hex())
        # print(sig.signature.hex())







import requests
from web3.auto import w3
from eth_account.messages import encode_defunct, _hash_eip191_message
from jose.utils import base64url_encode

class Transaction:
    def __init__(self, tx_id, token_symbol, action, from_, to, amount, 
                    fee, fee_recipient, nonce, token_id, chain_type, chain_id, data, version):
        self.tx_id = tx_id
        self.token_symbol = token_symbol
        self.action = action
        self.from_ = from_
        self.to = to
        self.amount = amount
        self.fee = fee
        self.fee_recipient = fee_recipient
        self.nonce = nonce
        self.token_id = token_id
        self.chain_type = chain_type
        self.chain_id = chain_id
        self.data = data
        self.version = version
        self.sig = ''
        self.ever_hash = self.get_ever_hash()

    def __str__(self):
        return 'tokenSymbol:' + self.token_symbol + '\n' + \
                'action:' + self.action + '\n' + \
                'from:' + self.from_ + '\n' + \
                'to:' + self.to + '\n' + \
                'amount:' + self.amount + '\n' + \
                'fee:' + self.fee + '\n' + \
                'feeRecipient:' + self.fee_recipient + '\n' + \
                'nonce:' + self.nonce + '\n' + \
                'tokenID:' + self.token_id + '\n' + \
                'chainType:' + self.chain_type + '\n' + \
                'chainID:' + self.chain_id + '\n' + \
                'data:' + self.data + '\n' + \
                'version:' + self.version 

    def sign(self, private_key):
        pk = bytearray.fromhex(private_key)
        message = encode_defunct(text=str(self))
        sig = w3.eth.account.sign_message(message, private_key=pk)
        self.sig = sig.signature.hex()
        return sig
    
    def sign_with_ar_wallet(self, wallet):
        h = _hash_eip191_message(encode_defunct(text=str(self)))
        sig = base64url_encode(wallet.sign(h)).decode()
        self.sig = sig
        return sig

    def get_ever_hash(self):
        message = encode_defunct(text=str(self))
        message_hash = _hash_eip191_message(message)
        return w3.toHex(message_hash)

    def to_dict(self):
        return {
            'tx_id': self.tx_id,
            'tokenSymbol': self.token_symbol,
            'action': self.action,
            'from': self.from_,
            'to': self.to,
            'amount': self.amount,
            'fee': self.fee,
            'feeRecipient': self.fee_recipient,
            'nonce': self.nonce,
            'tokenID': self.token_id,
            'chainType': self.chain_type,
            'chainID': self.chain_id,
            'data': self.data,
            'version': self.version,
            'sig': self.sig
        }

    def post(self, api_host):
        url = api_host + '/tx'
        if not self.sig:
            raise Exception(self.tx_id, 'no signature.')
        #print('post_data:', self.to_dict())
        return requests.post(url, json=self.to_dict())


TOKEN_LIST = [
    #kovan
    ['ethereum', '42', 'eth', '0x0000000000000000000000000000000000000000', 18],
    ['ethereum', '42', 'usdt', '0xd85476c906b5301e8e9eb58d174a6f96b9dfc5ee', 6],

    #mainnet
    ['ethereum', '1', 'eth', '0x0000000000000000000000000000000000000000', 18],
    ['ethereum', '1', 'usdt', '0xdac17f958d2ee523a2206206994597c13d831ec7', 6],
    ['ethereum', '1', 'wbtc', '0x2260fac5e5542a773aa44fbcfedf7c193bc2c599', 8],
    ['ethereum', '1', 'usdc', '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48', 6],
    ['ethereum', '1', 'dai', '0x6b175474e89094c44da98b954eedeac495271d0f', 18],
    ['ethereum', '1', 'uni', '0x1f9840a85d5af5bf1d1762f925bdaddc4201f984', 18],
    ['ethereum', '1', 'sos', '0x3b484b82567a09e2588a13d54d032153f0c0aee0', 18]
]

def get_token_id(chain_type, chain_id, token_symbol):
    for token in TOKEN_LIST:
        if token[0] == chain_type and token[1] == chain_id and token[2] == token_symbol:
            return token[3]

def get_token_decimal(chain_type, chain_id, token_symbol):
    for token in TOKEN_LIST:
        if token[0] == chain_type and token[1] == chain_id and token[2] == token_symbol:
            return token[4]
    
def get_token_tag(chain_type, chain_id, token_symbol, token_id=''):
    if not token_id:
        token_id = get_token_id(chain_type, chain_id, token_symbol)

    return '-'.join([chain_type, token_symbol, token_id])